/***************************************************************************
 *                                                                         *
 *   Copyright ©2020 KylinSec. All rights reserved.                      *
 *                                                                         *
 ***************************************************************************/
#include "kylinsec-security-center-dbus.h"
#include "kylinsec-security-center-dbus_adaptor.h"
#include <QDBusConnection>
#include <QDBusError>
#include <QSettings>
#include <QProcess>
#include <QFile>
#include <QJsonDocument>
#include <QDir>
#include <QDebug>

KylinsecSecurityCenterDbus::KylinsecSecurityCenterDbus(QObject *parent) : QObject(parent)
{
    new SfAdaptor(this);
    QDBusConnection connection = QDBusConnection::systemBus();
    if (!connection.isConnected()) {
        qWarning("Cannot connect to the D-Bus session bus.\n"
                 "Please check your system settings and try again.\n");
    }

    if (!connection.registerObject(KSCMMS_DBUS_PATH_NAME, this))
    {
        qWarning() << "error:" << connection.lastError().message();
    }

    if(!connection.registerService(KSCMMS_DBUS_SERVICE_NAME))
    {
        qWarning() << "error:" << connection.lastError().message();
    }
    GetSpecialTag = isSpecial();
}

int KylinsecSecurityCenterDbus::setLoginFailedOpr(int failCount, int lockTime)
{
    if (1 == GetSpecialTag) {
        QStringList files;
        files << "/etc/pam.d/login" << "/etc/pam.d/lightdm" << "/etc/pam.d/sshd";
        QString replaceStr = (failCount==0) ? QString() : QString("auth       required     pam_tally2.so deny=%1 unlock_time=%2 even_deny_root root_unlock_time=%2").arg(failCount).arg(lockTime);
        foreach (QString file, files) {
            if (setSettingsValue(file, "^auth +.*pam_tally2.so", replaceStr, "auth") != 0)
                return 1;
        }

        return 0;
    }

    QStringList files;
    files << "/etc/pam.d/system-auth" << "/etc/pam.d/password-auth";
    foreach (QString file, files) {
        QString replaceStr = (failCount==0) ? QString() : QString("auth        requisite                                    pam_faillock.so preauth audit deny=%1 even_deny_root unlock_time=%2")
                                              .arg(failCount).arg(lockTime);
        int ret = setSettingsValue(file, "^auth +.*pam_faillock.so +preauth", replaceStr, "auth", "^auth +.*pam_env.so", true);
        if(ret != 0) return 1;

        replaceStr = (failCount==0) ? QString() : QString("auth        [default=die]                                pam_faillock.so authfail audit deny=%1 even_deny_root unlock_time=%2")
                                      .arg(failCount).arg(lockTime);
        ret = setSettingsValue(file, "^auth +.*pam_faillock.so +authfail", replaceStr, "auth", "^auth +.*pam_unix.so", true);
        if(ret != 0) return 1;

        replaceStr = (failCount==0) ? QString() : QString("auth        sufficient                                   pam_faillock.so authsucc audit deny=%1 even_deny_root unlock_time=%2")
                                      .arg(failCount).arg(lockTime);
        ret = setSettingsValue(file, "^auth +.*pam_faillock.so +authsucc", replaceStr, "auth", "^auth +.*pam_faillock.so +authfail", true);
        if(ret != 0)
            return 1;
    }
    return 0;
}

QString KylinsecSecurityCenterDbus::getLoginFailedOpr()
{
    if (1 != GetSpecialTag) {
        QStringList ret = getSettingsValue("/etc/pam.d/system-auth", "^auth +.*pam_faillock.so +.*deny *= *(\\d+) +.*unlock_time *= *(\\d+)");
        return (ret.count() < 2) ? QString("{\"failCount\": 0, \"lockTime\": 0}") : QString("{\"failCount\": %1, \"lockTime\": %2}").arg(ret.at(0)).arg(ret.at(1));
    } else {
        QStringList ret = getSettingsValue("/etc/pam.d/system-auth", "^auth +.*pam_tally2.so +.*deny *= *(\\d+) +.*unlock_time *= *(\\d+)");
        return (ret.count() < 2) ? QString("{\"failCount\": 0, \"lockTime\": 0}") : QString("{\"failCount\": %1, \"lockTime\": %2}").arg(ret.at(0)).arg(ret.at(1));
    }
}

int KylinsecSecurityCenterDbus::setPasswordComplex(bool enabled)
{
    QString replaceStr = enabled ? QString("password    requisite                                    pam_pwquality.so try_first_pass local_users_only enforce_for_root minlen=8 dcredit=-1 lcredit=-1 ocredit=-1 retry=1 type=")
                                 : QString("password    requisite                                    pam_pwquality.so") ;
    return setSettingsValue("/etc/pam.d/system-auth", "^password +.*pam_pwquality.so", replaceStr, "password");
}

bool KylinsecSecurityCenterDbus::getPasswordComplex()
{
    QStringList ret = getSettingsValue("/etc/pam.d/system-auth", "^password +.*pam_pwquality.so +try_first_pass +local_users_only +enforce_for_root +minlen *= *(\\d+) +dcredit *= *-1 +lcredit *= *-1 +ocredit= *-1 +retry *= *1 +type *=");
    return ret.count() == 1;
}
/**
 * @brief KylinsecSecurityCenterDbus::setPasswordExpired  /etc/shadow（影子文件）所有伪用户的密码都是 "!!" 或 "*"，代表没有密码是不能登录的。当然，新创建的用户如果不设定密码，那么它的密码项也是 "!!"，代表这个用户没有密码，不能登录。
 * @param days
 * @return
 */
int KylinsecSecurityCenterDbus::setPasswordExpired(int days)
{
    QFile file("/etc/shadow");
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return 1;

    QStringList loginNames;
    while (!file.atEnd()) {
        QString line = QString(file.readLine());
        QStringList values = line.split(":");
        if(values.count() != 9) continue;

        if(values.at(1) != "*" && values.at(1) != "!!")
            loginNames << values.at(0);
    }

    foreach (QString name, loginNames) {
        QString cmd = QString("chage -M %1 %2").arg(days).arg(name);
        int ret = QProcess::execute(cmd);
        if(ret != 0) return 1;
    }

     QString replaceStr = QString("PASS_MAX_DAYS	%1").arg(days);
     return setSettingsValue("/etc/login.defs", "^PASS_MAX_DAYS", replaceStr);
}

int KylinsecSecurityCenterDbus::getPasswordExpired()
{
    QFile file("/etc/shadow");
    if (!file.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        qWarning() << file.errorString();
        return -1;
    }

    while (!file.atEnd()) {
        QString str(file.readLine());
        if(str.startsWith("root:"))
        {
            QStringList l = str.split(":", QString::KeepEmptyParts);
            if(l.count() >= 5)
            {
                return l.at(4).toInt();
                break;
            }
        }
    }

    file.close();

    return -1;
}

int KylinsecSecurityCenterDbus::setSelinuxConfig(QString config)
{
    QString replaceStr = config.isEmpty() ? QString() : QString("SELINUX=%1").arg(config);
    return setSettingsValue("/etc/selinux/config", "^SELINUX *=", replaceStr);
}

QString KylinsecSecurityCenterDbus::getSelinuxConfig()
{
    QStringList ret = getSettingsValue("/etc/selinux/config", "^SELINUX *= *(disabled|permissive|enforcing)");
    return ret.isEmpty() ? QString() : ret.first();
}

int KylinsecSecurityCenterDbus::setSessionTimeout(int seconds)
{
    QString replaceStr = (seconds == 0) ? QString() : QString("TMOUT=%1;export TMOUT").arg(seconds);
    return setSettingsValue("/etc/profile", "^TMOUT *= *(\\d+) *", replaceStr);
}

int KylinsecSecurityCenterDbus::getSessionTimeout()
{
    QStringList ret = getSettingsValue("/etc/profile", "^TMOUT *= *(\\d+) *; *export +TMOUT");
    return ret.isEmpty() ? 0 : ret.first().toInt();
}

int KylinsecSecurityCenterDbus::setAuthType(int type)
{
    Q_UNUSED(type)
    return 0;
}

int KylinsecSecurityCenterDbus::getAuthType()
{
    return 0;
}

bool KylinsecSecurityCenterDbus::copyFile(const QString &sourceFile, const QString &targetFile)
{
    if (QFile::exists(targetFile)){
        if(!QFile::remove(targetFile))
        {
            qWarning() << "remove file error:" << targetFile;
            return false;
        }
    }
    return QFile::copy(sourceFile, targetFile);
}
/**
 * @brief KylinsecSecurityCenterDbus::setSettingsValue
 * @param configFile
 * @param regexp 替换时，要替换的行的正则表达式
 * @param replaceStr 替换的字符串
 * @param insertPosRegexp 新增时，插入位置行的正则表达式
 * @param offset 新增时，插入位置的偏移量。
 * @return
 */
int KylinsecSecurityCenterDbus::setSettingsValue(const QString &configFile, const QString &regexp, const QString &replaceStr, const QString &groupName, const QString&insertRowRegexp, const bool &insertNextRow)
{
    if(configFile.isEmpty()) return 1;

    if(!copyFile(configFile, configFile+".bak"))
    {
        qWarning() << "bak file error.";
        return 1;
    }
    // read from file
    QFile file(configFile);
    if (!file.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        qWarning() << file.errorString();
        return 1;
    }
    QStringList strlist = QString(file.readAll()).split("\n", QString::KeepEmptyParts);
    file.close();
    int count = strlist.count();
    //change value
    QRegExp rx(regexp);
    int pos = 0;
    int groupStart = -1;
    int insertPos = -1;
    for(int i=0; i<count; ++i)
    {
        pos = 0;
        QString str = strlist.at(i).trimmed();
        if ((pos = rx.indexIn(str, 0)) != -1)
        {
            if(replaceStr.isEmpty())
            {
                strlist.removeAt(i);
            }
            else
            {
                strlist.replace(i, replaceStr);
            }
            break;
        }
        //如果需要归并到group的位置，则标记所属group的开始位置
        if(!groupName.isEmpty() && groupStart < 0)
        {
            if(str.startsWith(groupName)) groupStart = i;
        }

        if(!insertRowRegexp.isEmpty() && insertPos < 0)
        {
            QRegExp insert_pos_rx(insertRowRegexp);
            if ((pos = insert_pos_rx.indexIn(str, 0)) != -1)
            {
                insertPos = i + (insertNextRow ? 1 : 0);
            }
        }
    }
    //没有匹配到，需要添加。
    if(pos == -1)
    {
        if(insertPos >= 0)
        {
            strlist.insert(insertPos, replaceStr);
        }
        else
        {
            //如果需要归并到group的位置，默认插入到group的首行。
            if(groupStart > 0)
            {
                strlist.insert(groupStart, replaceStr);
            }
            else //否则插入到文件的最后一行。
                strlist.append(replaceStr);
        }
    }
    //write to file
    if (!file.open(QIODevice::ReadWrite | QIODevice::Text | QIODevice::Truncate))
    {
        qWarning() << file.errorString();
        return 1;
    }

    file.write(strlist.join("\n").toLatin1());
    file.close();

    return 0;
}

QStringList KylinsecSecurityCenterDbus::getSettingsValue(const QString &configFile, const QString &regexp)
{
    if(configFile.isEmpty()) return QStringList();

    // read from file
    QFile file(configFile);
    if (!file.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        qWarning() << file.errorString();
        return QStringList();
    }
    QStringList strlist = QString(file.readAll()).split("\n", QString::KeepEmptyParts);
    file.close();
    int count = strlist.count();
    //read value
    QRegExp rx(regexp);
    int pos = 0;
    for(int i=0; i<count; ++i)
    {
        pos = 0;
        QString str = strlist.at(i).trimmed();
        if ((pos = rx.indexIn(str, 0)) != -1)
        {
            QStringList ret = rx.capturedTexts();
            ret.removeFirst();
            return ret;
        }
    }

    return QStringList();
}

int KylinsecSecurityCenterDbus::isSpecial()
{
    FILE * fp = fopen("/etc/.kyinfo", "r");
    if (!fp) {
        perror("fopen err");
        return -1;
    }

    int ret = 0;
    do {
        fseek(fp, 0, SEEK_END);
        long filelen = ftell(fp);
        if (filelen < 0)
            break;

        fseek(fp, 0, SEEK_SET);
        std::vector<char> readbuf(filelen, 0);
        while(fgets(readbuf.data(), readbuf.size(), fp) != NULL) {
            if (strstr(readbuf.data(), "milestone")) {
                if (strstr(readbuf.data(), "3.3-6B")) {
                    ret = 1;
                }
                break;
            }
        }

    } while (0);


    if (fp)
        fclose(fp);
    fp = NULL;
    return ret;
}
